#include <conf.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <version.h>

struct CONF_T
{
  int      lower;
  int      upper;
  int      timeout;
  BOOLEAN  verbose;
  BOOLEAN  debug;
  char     *address;
  char     *rcfile;
  SCANTYPE scantype;
  PROTOCOL protocol;
  ARRAY    ports;
};

private void    __set_rcfile(CONF this);
private BOOLEAN __file_exists(char *file);
private char *  __get_line(FILE *fp);
private char *  __chomp_line(FILE *fp, char **str, int *line_num);
private BOOLEAN __strmatch(char *option, char *param);
private char *  __get_port_string(CONF this);

CONF
new_conf()
{
  CONF this;
  int i;
  int ports[] = {7, 21, 22, 23, 25, 37, 80, 110, 387, 443};

  this = calloc(sizeof(*this),1);
  this->verbose = TRUE;
  set_timeout(this, 1);
  set_lower(this, 1);
  set_upper(this, 1024);
  set_debug(this,FALSE);
  set_type(this, PORT);
  set_protocol(this, ICMP, FALSE);
  __set_rcfile(this); 

  this->ports = new_array();
  for(i = 0; i < 10; i++)
    array_put(this->ports, array_get_size(this->ports), (void*)ports[i]);
  return this;
}

void
conf_destroy(CONF this) {
  xfree(this->address);
  xfree(this->rcfile);
  xfree(this);
  return;
}

void
show(CONF this, BOOLEAN quit){
  printf("GENERAL DICK CONFIG\n");
  printf("  version:                      %s\n", version_string);
  printf("  verbose:                      %s\n", this->verbose?"true":"false");
  printf("  debug:                        %s\n", this->debug?"true":"false");
  printf("  socket timeout:               %d\n", this->timeout);
  printf("  resource file:                %s\n", this->rcfile);
  printf( "\n" );
  printf("NETSCANS                          \n");
  printf("  protocol:                     %s\n", this->protocol==ICMP?"ICMP":"TCP/IP");
  printf("  ports to scan (tcp only)      %s\n", __get_port_string(this));
  printf( "\n" );
  printf("PORTSCANS                          \n");
  printf("  lower port:                   %d\n", this->lower);
  printf("  upper port:                   %d\n", this->upper);
  printf( "\n" );
  if(quit) exit(0);
  return;
}

void
set_rcfile(CONF this, char *file){
  size_t len = strlen(file)+1;
  xrealloc(this->rcfile, len);
  snprintf(this->rcfile, len, "%s", file);
  return;
}

void
set_lower(CONF this, int i){
  if(i < 1){
    i = 1;
  }
  this->lower = i;
  return; 
}

void
set_upper(CONF this, int i){
  this->upper = i;
  return; 
}

void
set_verbose(CONF this, BOOLEAN b){
  this->verbose = b;
  return;
}

void
set_timeout(CONF this, int timeout){
  this->timeout = timeout;
  return;
}

void 
set_address(CONF this, char *address){
  this->address = xstrdup(address);
  return;
}

void
set_type(CONF this, SCANTYPE scantype){
  this->scantype = scantype;
  return;
}

void
set_protocol(CONF this, PROTOCOL protocol, BOOLEAN warn){
  int  uid = geteuid();
  if(uid != 0 && protocol==ICMP){
    if(warn) joe_error("You must be root to use ICMP, switching to TCP protocol\n");
    this->protocol = TCP;
  } else {
    this->protocol = protocol;
  }
  return;
}

void 
set_debug(CONF this, BOOLEAN debug){
  this->debug = debug;
}

char *
get_rcfile(CONF this){
  return this->rcfile;
}

char *
get_address(CONF this){
  return this->address;
}

int
get_lower(CONF this){
  return this->lower;
}

int 
get_upper(CONF this){
  return this->upper;
}

int
get_timeout(CONF this){
  return this->timeout;
}

SCANTYPE
get_type(CONF this){
  return this->scantype;
}

PROTOCOL
get_protocol(CONF this){
  return this->protocol;
}

BOOLEAN 
get_debug(CONF this){
  return this->debug;
}

ARRAY
get_ports(CONF this){
  return this->ports;
}

private char *
__get_port_string(CONF this){
  int   i;
  int   len = 0;
  char  *ptr;
  ARRAY A = get_ports(this);
  for(i = 0; i < (int)array_get_size(A); i++){
    char tmp[32];
    snprintf(tmp, 32, "%d", (int)array_get(A, i));
    len += strlen(tmp)+1;
  }
  ptr = xmalloc(len);
  memset(ptr, '\0', len);
  for(i = 0; i < (int)array_get_size(A); i++){
    char tmp[32];
    snprintf(tmp, 32, "%d", (int)array_get(A, i));
    strncat(ptr, tmp, len);
    if(i < (int)array_get_size(A)-1){
      strncat(ptr, ",", len);
    }
  }
  return ptr;
}

private void
__set_rcfile(CONF this){
  char *env;
  char *home;
  int  len;
 
  env = getenv("DICKRC");
  if(env && *env){
    if(!__file_exists(env)){
      fprintf(stderr, "ERROR: DICKRC points to '%s' which does not exist\n", env);
    } else {
      this->rcfile = xstrdup(env);
      return;
    }
  }

  home = getenv("HOME");
  if(!home){
    struct passwd *pwd = getpwuid (getuid ());
    if(!pwd || !pwd->pw_dir){
      return; /* we're getting nowhere...*/
    } else {
      home = pwd->pw_dir;
    }
  } 
  len  = strlen(home) + strlen(program_name) + 5;
  this->rcfile = xmalloc(len);
  snprintf(this->rcfile, len, "%s/.%src", home, program_name);
  return;
}

private BOOLEAN
__file_exists(char *file)
{
  int  fd;

  /* open the file read only  */
  if((fd = open(file, O_RDONLY)) < 0){
  /* the file does NOT exist  */
    close(fd);
    return FALSE;
  } else {
  /* party on Garth... */
    close(fd);
    return TRUE;
  }

  return FALSE;
}

private char *
__chomp_line(FILE *fp, char **str, int *line_num)
{
  char *ptr;
  while(TRUE){
    if((*str = __get_line(fp)) == NULL) return NULL;
    (*line_num)++;
    ptr = chomp(*str);
    /* exclude comments */
    if(*ptr != '#' && *ptr != '\0'){
      return(ptr);
    } else {
      xfree(ptr);
    }
  }
}

private char *
__get_line(FILE *fp) {
  char *ptr = NULL;
  char *newline;
  char tmp[256];

  memset(tmp, 0, sizeof(tmp));
  do {
    if((fgets(tmp, sizeof(tmp), fp)) == NULL) return(NULL);
    if(ptr == NULL){ 
      ptr = xstrdup(tmp);
    } else { 
      ptr = (char*)xrealloc(ptr, strlen(ptr) + strlen(tmp) + 1);
      strcat(ptr, tmp);
    }
    newline = strchr(ptr, '\n');
  } while(newline == NULL);
  *newline = '\0';
  
  return ptr;
}

private BOOLEAN
__strmatch(char *option, char *param) {
  if(!strncasecmp(option,param,strlen(param))&&strlen(option)==strlen(param))
    return TRUE;
  else
    return FALSE;
}

BOOLEAN
parse_rcfile(CONF this) {
  FILE *fp;
  int  line_num = 0;
  char *line;
  char *tmp;
  char *option;
  char *optionptr;
  char *value;

  if(!__file_exists(this->rcfile)){
    return FALSE;
  }

  if ((fp = fopen(this->rcfile, "r")) == NULL) {
    return FALSE;
  }

  while((line = __chomp_line(fp, &line, &line_num)) != NULL){
    tmp = line;
    optionptr = option = xstrdup(line);
    while(*optionptr && !ISSPACE((int)*optionptr) && !ISSEPARATOR(*optionptr))
      optionptr++;
    *optionptr++=0;
    while(ISSPACE((int)*optionptr) || ISSEPARATOR(*optionptr))
      optionptr++;
    value  = xstrdup(optionptr);
    while(*line)
      line++;
    if(__strmatch(option, "debug")){
      if(!strncasecmp(value, "true", 4)){
        this->debug = TRUE;
      } else {
        this->debug = FALSE;
      }
    }
    if(__strmatch(option, "verbose")){
      if(!strncasecmp(value, "true", 4)){
        this->verbose = TRUE;
      } else {
        this->verbose = FALSE;
      }
    }
    if(__strmatch(option, "timeout")){
      if(value && *value){
        this->timeout = atoi(value);
      } else {
        this->timeout = 1;
      }
    }
    if(__strmatch(option, "protocol")){
      int  uid = geteuid();
      if(!strncasecmp(value, "icmp", 4) && uid==0){
        this->protocol = ICMP;
      } else {
        this->protocol = TCP;
      }
    }
    if(__strmatch(option, "ports")){
      if(value && *value){
        char *comma = ",";
        char *next  = 0;
        this->ports = array_destroy(this->ports);
        this->ports = new_array();
        for(next = strtok(value, comma); next; next = strtok(0, comma)){
          int tmp;
          trim(next);
          tmp = atoi(next);
          array_put(this->ports, array_get_size(this->ports), (void*)tmp);
        }
      }
    }
    if(__strmatch(option, "lower")){
      if(value != NULL){
        this->lower = atoi(value);
      } else {
        this->lower = 1;
      }
    }
    if(__strmatch(option, "upper")){
      if(value != NULL){
        this->upper = atoi(value);
      } else {
        this->upper = 1;
      }
    }
    xfree(value);
    xfree(option);
    free(tmp);
  } /* end of while chomp_line */
        
  fclose(fp);
  return 0;
}  

